#include <iostream>
#include <cstdlib>
#include <cassert>
#include "mpi.h"

using namespace std;

// randomly generate a vector
double *genVector(const int N, const int seed)
{
    int i;
    double *A;
   
		A = new (nothrow) double[N]; 
    assert(A);
    
    srand(seed);    
    for(i=0;i<N;i++) {
        A[i] = ((double) rand() / RAND_MAX);
    }
    
    return A;
}

// get a vector size either from command line or from console
int getSize(int argc, char **argv)
{
    int length;
    // Determine the vector size
    if(argc>1) {
        length = atoi(argv[1]);
    }
    while(length<=1) {
				cout << "\nPlease enter the length of the vector: ";
				cin >> length;
    }    
    
    return length;
}

// Calculate inner product of two vectors, version 1
int main(int argc, char **argv) 
{
    int rank, size;
    double *v1=NULL, *v2=NULL;
    int length = 0;
    int i, myHead, myTail, myShare;
    double sum, allSum;
    double t1, t2;
    
    // Initialization
    MPI_Init(&argc, &argv);
    
    // Get environment information
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if(rank==0) {
        length = getSize(argc, argv);
        v1 = genVector(length, 100);
        v2 = genVector(length, 200);
    }    

    t1 = MPI_Wtime();
    
    // Broadcast these vectors to all nodes
    MPI_Bcast(&length, 1, MPI_INT, 0, MPI_COMM_WORLD);

    // Slaves do not have memory for vectors yet! 
    if(rank!=0) {
			  v1 = new (nothrow) double[length];
	  		v2 = new (nothrow) double[length];
 	      assert(v1);
 	      assert(v2);
    }

		// Broadcast vectors to all nodes
    MPI_Bcast(v1, length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(v2, length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    
    // Now each node determines which part they work on
    myShare = length / size;
    myHead = myShare * rank;
    myTail = myHead + myShare;
    if(rank == (size - 1)) {
        myTail = length;
    }
    
    // now perform the actual computation
    sum = 0.0;
    for(i=myHead; i<myTail; i++) {
        sum += v1[i] * v2[i];
    }
        
    // now send results back to the root
    MPI_Reduce(&sum, &allSum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    t2 = MPI_Wtime() - t1;
    if(rank==0) {
        cout << "\nThe inner dot: " << allSum << ", " << t2 << " seconds.";
    }

    // Finalize
    MPI_Finalize();
    
    delete []v1;
    delete []v2;
    
    return 0;
}
