#include <cassert>
#include <cmath>
#include <cstdlib>
#include <iostream>

#include "mpi.h"
#define maxNP 32
using namespace std;

// randomly generate a vector
double *genVector(const int N, const int seed)
{
    int i;
    double *A;
    
		A = new double[N];
    assert(A);
    
    srand(seed);    
    for(i=0;i<N;i++) {
        A[i] = ((double) rand() / RAND_MAX);
    }
    
    return A;
}

// get a vector size either from command line or from console
int getSize(int argc, char **argv)
{
    int length=0;
    // Determine the vector size
    if(argc>1) {
        length = atoi(argv[1]);
    }
    while(length<=1) {
        cout << "\nPlease enter the length of the vector: ";
				cin >> length;
    }    
    
    return length;
}

// Calculate inner product of two vectors, version 2
int main(int argc, char **argv) 
{
    int rank, size;
    double *v1=NULL;
    int length = 0;
    double t1, t2;
    int work;
    int count[maxNP], displ[maxNP];
    
    int mySize, i;
    double *myData;
    double localSum, vLength;

    // Initialization
    MPI_Init(&argc, &argv);
    
    // Get environment information
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if(size > maxNP) {
        if(rank==0) {
            cout << "\nThis program is not capable of using more than " << maxNP << " processes.";
        }
				MPI_Finalize();
        return 255;
    }
    
    if(rank==0) {
        length = getSize(argc, argv);
        v1 = genVector(length, 100);
        
        // determine how work is going to be distributed
        work = length / size;
        displ[0] = 0;
        for(i=0;i<size-1;i++) {
            count[i] = work;
            displ[i+1] = displ[i] + work;
        }
        count[i] = length - (size-1) * work;
    }    

    t1 = MPI_Wtime();

    // let everyone knows how much work they have to do
    MPI_Scatter(count, 1, MPI_INT, &mySize, 1, MPI_INT, 0, MPI_COMM_WORLD);

    // every one allocate memory        
		myData = new double[mySize];

    // I am assuming memory allocation will success, which is not good!!!
    MPI_Scatterv(v1, count, displ, MPI_DOUBLE, myData, mySize, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    
    // now every node does computation
    localSum = 0.0;
    for(i=0;i<mySize;i++) {
        localSum += myData[i] * myData[i];
    }

    // sum all squres
    MPI_Allreduce(&localSum, &vLength, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    vLength = sqrt(vLength);

    if(vLength != 0.0) {
        // Now normalize the vector
        for(i=0;i<mySize;i++) {
            myData[i] /= vLength;
        }
    }        

    // now send results back to the root
    MPI_Gatherv(myData, mySize, MPI_DOUBLE, v1, count, displ, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    t2 = MPI_Wtime() - t1;
    if(rank==0) {
        cout << "\nNormalization takes " << t2 << " seconds.";
    }

    // Finalize
    MPI_Finalize();
    
    delete []v1;
    
    return 0;
}
